#include <RF24.h>
#include "emergency.h"

RF24 _radio(7, 8);                 // nRF24L01 radio on SPI bus + pins 7, 8
const byte _nodeId = 7;            // ID of this node
bool _verbose = false;             // enable/disable verbose messages
int _lastMsgSent[MAX_ID];          // ID of last message sent from each source
int _timesSeen[MAX_ID];            // number of times lastMsg has been seen
int _counter = 1;                  // message counter for source nodes
char _opMode = 'N';                // mode: N none, E emergency, C civilian
struct Location _location;         // last known location of this node

void setup() {
  Serial.begin(115200);
  setupRadio(_radio);
  _radio.startListening();
  //_location.latitude = 39.3243421;
  //_location.longitude = -76.6230005;
}

void propagateMsg(EmergencyMsg& eMsg) {
  if (eMsg.ttl == 0)
    return;
  /* Prepare message */
  eMsg.ttl--;
  eMsg.senderPath[MAX_TTL-eMsg.ttl-1] = _nodeId;
  eMsg.senderId = _nodeId;
  eMsg.senderLoc = _location;
  /* Wait for channel to be quiet */
  int tries = 0;
  while (_radio.testCarrier()) {        // if other nodes communicating,
    delayMicroseconds(random(10000));   //  wait and try a little later
    _radio.stopListening();             // reset flag for testCarrier()
    _radio.startListening();
    if (tries++ > 10)                   // don't send at all if too busy
      return;
  }
  /* Write out updated message */
  _radio.stopListening();
  _radio.write(&eMsg, sizeof(EmergencyMsg), 1);
  _radio.startListening();
}

void civilianLoop() {
  /* Receive and propagate messages */
  EmergencyMsg eMsg;
  if (_radio.available()) {
    _radio.read(&eMsg, sizeof(EmergencyMsg));
    eMsg.serialWriteLoc(_nodeId);
    /* Decide whether or not to propagate */
    if (eMsg.msgId > _lastMsgSent[eMsg.srcId]) {
      _lastMsgSent[eMsg.srcId] = eMsg.msgId;
      _timesSeen[eMsg.srcId] = 0;
      propagateMsg(eMsg);   // always propagate new messages
    } else if (eMsg.msgId < _lastMsgSent[eMsg.srcId]) {
      return;               // never propagate old messages
    } else if (random(int(pow(2, _timesSeen[eMsg.srcId]))) == 0) {
      propagateMsg(eMsg);   // decrease probability to send each time we send
    }                       //  chance to send: 1, 1/2, 1/4, 1/8, ...
    _timesSeen[eMsg.srcId]++;
  }
}

void emergencyLoop() {
  /* Initialize and send out emergency message */
  EmergencyMsg eMsg(_nodeId, _counter);
  eMsg.srcLoc = _location;
  _radio.stopListening();
  _radio.write(&eMsg, sizeof(EmergencyMsg), 1);
  _counter++;
  /* Get propagated messages, update android w/ locations */
  unsigned long startTime = micros();
  _radio.startListening();
  while (micros() - startTime < 2000000) {    // 2 sec between messages
    if (_radio.available()) {
      _radio.read(&eMsg, sizeof(EmergencyMsg));
      eMsg.serialWriteLoc(_nodeId);
      if (_verbose) {
        Serial.print("Got response in ");
        Serial.print(micros() - startTime);
        Serial.print(" us. ");
        eMsg.printMsg();
      }
    }
  }
}
/* Update this node's operation mode from android */
void updateOpMode() {
  if (Serial.available()) {
    char c = Serial.peek();
    if (c == 'C' || c == 'E' || c == 'N') {
      _opMode = Serial.read();
    } else if (c == 'V') {
      Serial.read();
      _verbose = true;
    }
  }
}
/* Update this node's location information from android */
void updateLocation() {
  if (Serial.available()) {
    char c = Serial.peek();
    if (isDigit(c) || c == '-') {
      _location.latitude = Serial.parseFloat();
      Serial.read();    // skip comma
      _location.longitude = Serial.parseFloat();
      Serial.read();    // skip semicolon
    }
  }
}

void loop() {
  updateOpMode();
  updateLocation();
  if (_opMode == 'C')
    civilianLoop();
  else if (_opMode == 'E')
    emergencyLoop();
}
