package com.rkinney4.httptest;

import android.content.Context;
import android.util.Log;
import android.widget.TextView;

import com.github.mikephil.charting.charts.CombinedChart;
import com.github.mikephil.charting.components.Legend;
import com.github.mikephil.charting.data.ChartData;
import com.github.mikephil.charting.data.CombinedData;
import com.github.mikephil.charting.data.DataSet;
import com.github.mikephil.charting.data.Entry;
import com.github.mikephil.charting.data.LineData;
import com.github.mikephil.charting.data.LineDataSet;
import com.github.mikephil.charting.data.ScatterData;
import com.github.mikephil.charting.data.ScatterDataSet;
import com.github.mikephil.charting.utils.EntryXComparator;

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

/**
 * Created by Rachel on 3/2/2017.
 */

public class Chart {

    private static final int LINE = 0;
    private static final int SCATTER = 1;

    private Context context;

    private ArrayList<DataPoint> data;
    private DataPoint.Field xField, yField, sField; // x, y, and series data fields
    private HashMap<String, Integer> series;

    private CombinedChart chart;
    private CombinedData cData;
    private ChartData[] chartData;

    private TextView title;
    private String titleStr;
    private int style, colorIndex;
    private int[] colors;

    Chart (Context context, CombinedChart chart, DataPoint.Field xField, DataPoint.Field yField, DataPoint.Field sField) {
        this.context = context;
        this.chart = chart;

        data = new ArrayList<>();
        series = new HashMap<>();
        colorIndex = -1;
        colors = context.getResources().getIntArray(R.array.seriesColors);

        // Create Data
        cData = new CombinedData();
        chartData = new ChartData[2];
        chartData[LINE] = new LineData();
        chartData[SCATTER] = new MyScatterData();
        cData.setData((LineData) chartData[LINE]);
        cData.setData((ScatterData) chartData[SCATTER]);
        chart.setData(cData);

        // Chart Visuals
        chart.getDescription().setEnabled(false);
        chart.setDrawGridBackground(true);
        chart.setDrawBorders(true);
        title = (TextView) ((MainActivity) context).findViewById(R.id.chart_title);
        titleStr = context.getString(R.string.chart_title);
        Legend leg = chart.getLegend();
        leg.setWordWrapEnabled(true);

        // Initialize
        updateFields(xField, yField, sField);
    }

    public void addPoint(DataPoint point) {
        data.add(point);
        String key = point.fieldToString(sField);
        if (!series.containsKey(key)) {
            addSeries(key);
        }
        if (style == LINE)
            chartData[style].addEntry(new Entry(point.getNumField(xField), point.getNumField(yField)), series.get(key));
        else
            ((MyScatterData)chartData[style]).addEntry(new Entry(point.getNumField(xField), point.getNumField(yField)), series.get(key));
        Log.i("addPoint", "Added point: (" + point.getNumField(xField) + ", " + point.getNumField(yField) + ")");
        refresh();
    }

    public void refresh() {
        cData.notifyDataChanged();
        chart.notifyDataSetChanged();
        chart.invalidate();
    }

    public void reset() {
        series.clear();
        data.clear();
        for (ChartData c : chartData)
            c.clearValues();
        cData.clearValues();
        chart.clearValues();
        colorIndex = -1;
        refresh();
    }

    public void updateFields(DataPoint.Field xField, DataPoint.Field yField, DataPoint.Field sField) {
        this.xField = xField;
        this.yField = yField;
        this.sField = sField;

        switch (xField) {
            case ID : style = LINE;
                break;
            default : style = SCATTER;
        }

        series.clear();
        for (ChartData c : chartData)
            c.clearValues();
        cData.clearValues();
        chart.clearValues();
        colorIndex = -1;
        initSeries();
        updateTitle();
        refresh();
        Log.i("updateFields", "Fields updated: x: " + ((MainActivity)context).fieldTypeToString(xField) +
                                            ", y: " + ((MainActivity)context).fieldTypeToString(yField) +
                                            ", s: " + ((MainActivity)context).fieldTypeToString(sField));
    }

    public void updateFields(String xField, String yField, String sField) {
        updateFields(((MainActivity)context).stringToFieldType(xField),
                    ((MainActivity)context).stringToFieldType(yField),
                    ((MainActivity)context).stringToFieldType(sField));
    }

    public boolean writeData(ObjectOutputStream oos) {
        try {
            oos.writeObject(data);
            return true;
        } catch (Exception e) {
            Log.e("writeData", "Failed to write data: " + e.getMessage());
            e.printStackTrace();
            return false;
        }
    }

    public boolean loadData(ObjectInputStream ois) {
        ArrayList<DataPoint> d;
        try {
            d = (ArrayList<DataPoint>) ois.readObject();
        } catch (Exception e) {
            Log.e("loadData", "Failed to load data: " + e.getMessage());
            e.printStackTrace();
            return false;
        }
        reset();
        data = d;
        initSeries();
        refresh();
        return true;
    }

    /*
     * Private Helper Functions
     */

    private void initSeries() {
        HashMap<String, ArrayList<Entry>> points = new HashMap<>();
        for (DataPoint d : data) {
            String key = d.fieldToString(sField);
            if (!points.containsKey(key)) {
                points.put(key, new ArrayList<Entry>());
            }
            points.get(key).add(new Entry(d.getNumField(xField), d.getNumField(yField)));
        }

        ArrayList<Entry> list;
        for (String s : points.keySet()) {
            list = points.get(s);
            Collections.sort(list, new EntryXComparator());
            addSeries(s, list);
        }
    }

    private void addSeries(String key) {
        addSeries(key, null);
    }

    private void addSeries(String key, ArrayList<Entry> list) {
        DataSet d;
        int index;
        int color = colors[(++colorIndex) % colors.length];
        switch (style) {
            case LINE:
                if (list == null)
                    d = new LineDataSet(new ArrayList<Entry>(), key);
                else
                    d = new LineDataSet(list, key);
                d.setColor(color);
                ((LineDataSet) d).setCircleColor(color);
                ((LineData)chartData[LINE]).addDataSet((LineDataSet) d);
                index = ((LineData)chartData[LINE]).getIndexOfDataSet((LineDataSet) d);
                break;
            case SCATTER:
                if (list == null)
                    d = new MyScatterDataSet(new ArrayList<Entry>(), key);
                else
                    d = new MyScatterDataSet(list, key);
                d.setColor(color);
                /* TODO add Formatting */
                ((ScatterData)chartData[SCATTER]).addDataSet((ScatterDataSet) d);
                index = ((ScatterData)chartData[SCATTER]).getIndexOfDataSet((ScatterDataSet) d);
                break;
            default:
                Log.e("addSeries", "Unknown style");
                index = -1;
        }
        series.put(key, index);
        Log.i("addSeries", "Added series: " + key);
    }

    private void updateTitle() {
        title.setText(String.format(titleStr,
                                   ((MainActivity)context).fieldTypeToString(yField),
                                   ((MainActivity)context).fieldTypeToString(xField)));
    }

    // Fixes Scatter Ordering problem
    private class MyScatterData extends ScatterData {
        @Override
        public void addEntry(Entry e, int dataSetIndex) {
            if (mDataSets.size() > dataSetIndex && dataSetIndex >= 0) {
                MyScatterDataSet set = (MyScatterDataSet) mDataSets.get(dataSetIndex);
                set.addEntryOrdered(e); /* Changed from addEntry() to addEntryOrdered() */
                calcMinMax(e, set.getAxisDependency());
            } else {
                Log.e("addEntry", "Cannot add Entry because dataSetIndex too high or too low.");
            }
        }
    }

    // Fixes Scatter Ordering problem
    private class MyScatterDataSet extends ScatterDataSet {

        EntryXComparator xcomp;

        public MyScatterDataSet(List<Entry> yVals, String label) {
            super(yVals, label);
            xcomp = new EntryXComparator();
        }

        @Override
        public void addEntryOrdered(Entry e) {
            if (e == null)
                return;

            if (mValues == null) {
                mValues = new ArrayList<Entry>();
            }

            calcMinMax(e);

            if (mValues.size() > 0 && mValues.get(mValues.size() - 1).getX() > e.getX()) {
                /* Changed to a correct binary search */
                int closestIndex = Arrays.binarySearch(mValues.toArray(new Entry[mValues.size()]), e, xcomp);
                if (closestIndex < 0)
                    closestIndex = Math.abs(closestIndex+1);
                mValues.add(closestIndex, e);
            } else {
                mValues.add(e);
            }
        }
    }
}
