#!/usr/bin/env python3

from flask import Flask, jsonify, request, g
import os
from contextlib import closing
from gcm import GCM
from redis import StrictRedis
from datetime import datetime, timedelta, timezone
from util import RepeatedTimer

# https://github.com/geeknam/python-gcm
app = Flask(__name__)
fcm_url = 'https://fcm.googleapis.com/fcm/send'
api_key = os.environ['FCM_API_KEY']
# api_key is found in the console -> settings -> cloud messaging -> server key
gcm = GCM(api_key)
DISTANCE = 30 # meters
PRUNE_INTERVAL = 10 #seconds
CAR_TTL = timedelta(30) # how long to live
PRUNE_KEY = 'prune'
CARS_KEY = 'cars'
def build_data_source():
  redis_host = os.environ.get('REDIS_HOST', '127.0.0.1')
  redis_port = int(os.environ.get('REDIS_PORT', 6379))
  redis_db = int(os.environ.get('REDIS_DB', 1))
  redis_conn_pool = StrictRedis(host=redis_host, port=redis_port, db=redis_db)
  return redis_conn_pool

redis_instance = build_data_source()

def utc_now():
  return datetime.utcnow().replace(tzinfo=timezone.utc)

def prune():
  with redis_instance.pipeline() as pipe:
    pipe.watch(PRUNE_KEY)
    timestamp = int(utc_now().timestamp())
    expired_vids = pipe.zrange(PRUNE_KEY, 0, timestamp)
    print('pruned %d with val %d' % (len(expired_vids), timestamp))
    if len(expired_vids) > 0:
      pipe.multi()
      pipe.zrem(PRUNE_KEY, *expired_vids)
      num_removed = pipe.zrem(CARS_KEY, *expired_vids)
      #pipe.zremrangebyscore(CARS_KEY, 0, timestamp)
      pipe.execute()
      print('pruned: %d cars' % len(expired_vids))
  return

def touch_car(car_id):
  option = 'CH' #if update else 'NX'
  ttl = int((utc_now() + CAR_TTL).timestamp())
  redis_instance.execute_command('ZADD', PRUNE_KEY, option, ttl, car_id)

class Pruner:
  def __init__(self, data_source) -> None:
    self._data_source = data_source
  def start(self) -> None:
    RepeatedTimer(PRUNE_INTERVAL, prune).foo()

def update_car(lat, lon, token):
  print('LAT LON %f %f' % (lat, lon))
  redis_instance.geoadd("cars", lon, lat, token)
  print('updated car OK, now calling touch_vid')
  touch_car(token)
  print('touched car')
  return jsonify({'status' : 'updated_car'})

def get_nearby_vehicles(lat, lon, withcoord):
  return redis_instance.georadius("cars", lon, lat, DISTANCE, withcoord=withcoord)

def get_nearby_vehicles_map(lat, lon):
  to_return = get_nearby_vehicles(lat, lon, True)
  return_map = {}
  for item in to_return:
    return_map[item[0]] = item[1]
  to_return_map = {}
  to_return_map['vals'] = return_map
  return to_return_map

#Update normal car locations in Redis
@app.route('/car_loc', methods=('POST',))
def update_car_location_in_redis():
  data_point = request.get_json()
  print(data_point)
  device_id = data_point['id']
  latitude = data_point.get('lat', 0.0)
  longitude = data_point.get('lon', 0.0)
  # update_redis
  update_car(latitude, longitude, device_id)
  to_return_map = get_nearby_vehicles_map(latitude, longitude)
  return jsonify(to_return_map)

#Update emergency vehicle in Redis
@app.route('/emergency', methods=('POST',))
def emergency_ping_all():
  message_json = request.get_json()
  token = message_json['id']
  lat = message_json['lat']
  lon = message_json['lon']
  print("emergency at: %f, %f," % (lat, lon))
  #update_car(lat, lon, token)
  tokens = get_nearby_vehicles(lat, lon, False)

  data = {'e_id' : token, 'lat' : lat, 'lon' : lon} # TODO perhaps add location information
  if (len(tokens) > 0):
    response = gcm.json_request(registration_ids=tokens, data=data)
    print('%s' % response)
  to_return_map = get_nearby_vehicles_map(lat, lon)
  return jsonify(to_return_map)


@app.route('/', methods=('HEAD',))
def ping_response():
  return jsonify({})

if __name__ == '__main__':
  host = os.environ.get('FLASK_HOST', '127.0.0.1')
  port = int(os.environ.get('FLASK_PORT', 8888))
  print(host)
  debug = bool(int(os.environ.get('FLASK_DEBUG', True)))
  Pruner(redis_instance).start()
  app.run(host=host, port=port, debug=debug, processes=4)
